@@ -51,6 +51,7 @@
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
 #include <unistd.h>
+#include <sys/user.h>
 #include "cryptouser.h"
 
 #include "kcapi.h"
@@ -62,7 +63,7 @@
 #define MINVERSION 6 /* API compatible, ABI may change, functional
 		      * enhancements only, consumer can be left unchanged if
 		      * enhancements are not considered */
-#define PATCHLEVEL 1 /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 2 /* API / ABI compatible, no functional changes, no
 		      * enhancements, bug fixes only */
 
 /* remove once in if_alg.h */
@@ -72,9 +73,6 @@
 #ifndef ALG_SET_AEAD_AUTHSIZE
 #define ALG_SET_AEAD_AUTHSIZE		5
 #endif
-#ifndef ALG_SET_RNG_SEED
-#define ALG_SET_RNG_SEED		6
-#endif
 
 /* remove once in socket.h */
 #ifndef AF_ALG
@@ -84,6 +82,7 @@
 #define SOL_ALG 279
 #endif
 
+#define MAXPIPELEN (16 * PAGE_SIZE)
 
 /************************************************************
  * Internal logic
@@ -91,7 +90,7 @@
 
 static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 				       struct iovec *iov, size_t iovlen,
-				       uint32_t enc, int more)
+				       uint32_t enc, unsigned int flags)
 {
 	ssize_t ret = -EINVAL;
 	char *buffer = NULL;
@@ -159,7 +158,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 		*assoclen = handle->aead.assoclen;
 	}
 
-	ret = sendmsg(handle->opfd, &msg, more ? MSG_MORE : 0);
+	ret = sendmsg(handle->opfd, &msg, flags);
 
 	memset(buffer, 0, bufferlen);
 	/* magic to convince GCC to memset the buffer */
@@ -170,43 +169,66 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	return ret;
 }
 
-static ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
+static inline ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
 				       struct iovec *iov, size_t iovlen,
-				       int more)
+				       unsigned int flags)
 {
 	struct msghdr msg;
 
-	memset(&msg, 0, sizeof(msg));
+	msg.msg_name = NULL;
+	msg.msg_namelen = 0;
+	msg.msg_control = NULL;
+	msg.msg_controllen = 0;
+	msg.msg_flags = 0;
 	msg.msg_iov = iov;
 	msg.msg_iovlen = iovlen;
 
-	return sendmsg(handle->opfd, &msg, more ? MSG_MORE : 0);
+	return sendmsg(handle->opfd, &msg, flags);
 }
 
-static ssize_t _kcapi_common_vmsplice_data(struct kcapi_handle *handle,
-					   struct iovec *iov, size_t iovlen)
+static inline ssize_t _kcapi_common_vmsplice_data(struct kcapi_handle *handle,
+						  struct iovec *iov,
+						  size_t iovlen,
+						  size_t inlen,
+						  unsigned int flags)
 {
-	return vmsplice(handle->pipes[1], iov, iovlen, SPLICE_F_GIFT);
+	ssize_t ret = 0;
+
+	if (inlen > MAXPIPELEN) {
+		fprintf(stderr, "Splice operation: input data exceeds maximum allowed size of %lu\n",
+			MAXPIPELEN);
+		return -E2BIG;
+	}
+
+	ret = vmsplice(handle->pipes[1], iov, iovlen, SPLICE_F_GIFT|flags);
+	if (0 > ret)
+		return ret;
+	return splice(handle->pipes[0], NULL, handle->opfd, NULL, inlen, flags);
 }
 
-static ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
-				       struct iovec *iov, size_t iovlen)
+static inline ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
+					      struct iovec *iov, size_t iovlen)
 {
 	struct msghdr msg;
+	ssize_t ret = 0;
 
-	memset(&msg, 0, sizeof(msg));
+	msg.msg_name = NULL;
+	msg.msg_namelen = 0;
+	msg.msg_control = NULL;
+	msg.msg_controllen = 0;
+	msg.msg_flags = 0;
 	msg.msg_iov = iov;
 	msg.msg_iovlen = iovlen;
 
-	return recvmsg(handle->opfd, &msg, 0);
+	ret = recvmsg(handle->opfd, &msg, 0);
+	if (msg.msg_flags & MSG_TRUNC)
+		return -ENOMEM;
+	return ret;
 }
 
-static ssize_t _kcapi_common_splice_data(struct kcapi_handle *handle,
-					 size_t inlen,
-				         unsigned char *out, size_t outlen)
+static inline ssize_t _kcapi_common_read_data(struct kcapi_handle *handle,
+					      unsigned char *out, size_t outlen)
 {
-	if (0 > splice(handle->pipes[0], NULL, handle->opfd, NULL, inlen, 0))
-		return -EIO;
 	return read(handle->opfd, out, outlen);
 }
 
@@ -674,6 +696,7 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 {
 	struct iovec iov;
 	ssize_t ret = 0;
+	size_t processed = 0;
 	unsigned int bs = handle->info.blocksize;
 
 	if (!in || !inlen || !out || !outlen) {
@@ -691,7 +714,6 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 	}
 
 	iov.iov_base = (void*)(uintptr_t)in;
-	iov.iov_len = inlen;
 #if 0
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
@@ -699,6 +721,7 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 	 */
 	/* TODO make heuristic when one syscall is slower than four syscalls */
 	if (inlen < (1<<15)) {
+		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_ENCRYPT,
 					      0);
 		iov.iov_base = (void*)(uintptr_t)out;
@@ -706,14 +729,25 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 		return _kcapi_common_recv_data(handle, &iov, 1);
 }
 #endif
+
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT, 0);
 	if (0 > ret)
 		return ret;
 
-	ret = _kcapi_common_vmsplice_data(handle, &iov, 1);
-	if (0 > ret)
-		return ret;
-	return _kcapi_common_splice_data(handle, inlen, out, outlen);
+	while (inlen) {
+		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
+
+		iov.iov_len = datalen;
+		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
+						  SPLICE_F_MORE);
+		if (0 > ret)
+			return ret;
+		processed += ret;
+		iov.iov_base = (void*)(uintptr_t)(in + processed);
+		inlen -= ret;
+	}
+
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 
 /**
@@ -737,6 +771,7 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 {
 	struct iovec iov;
 	ssize_t ret = 0;
+	size_t processed = 0;
 
 	if (!in || !inlen || !out || !outlen) {
 		fprintf(stderr,
@@ -760,7 +795,6 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 	}
 
 	iov.iov_base = (void*)(uintptr_t)in;
-	iov.iov_len = inlen;
 #if 0
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
@@ -768,6 +802,7 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 	 */
 	/* TODO make heuristic when one syscall is slower than four syscalls */
 	if (inlen < (1<<15)) {
+		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_DECRYPT,
 					      0);
 		iov.iov_base = (void*)(uintptr_t)out;
@@ -775,14 +810,25 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 		return _kcapi_common_recv_data(handle, &iov, 1);
 	}
 #endif
+
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT, 0);
 	if (0 > ret)
 		return ret;
 
-	ret = _kcapi_common_vmsplice_data(handle, &iov, 1);
-	if (0 > ret)
-		return ret;
-	return _kcapi_common_splice_data(handle, inlen, out, outlen);
+	while (inlen) {
+		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
+
+		iov.iov_len = datalen;
+		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
+						  SPLICE_F_MORE);
+		if (0 > ret)
+			return ret;
+		processed += ret;
+		iov.iov_base = (void*)(uintptr_t)(in + processed);
+		inlen -= ret;
+	}
+
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 
 /**
@@ -813,7 +859,8 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 				     struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT, 1);
+	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT,
+				       MSG_MORE);
 }
 
 /**
@@ -844,7 +891,8 @@ ssize_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 ssize_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 				     struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT, 1);
+	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT,
+				       MSG_MORE);
 }
 
 /**
@@ -872,7 +920,7 @@ ssize_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 ssize_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
 				   struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_data(handle, iov, iovlen, 1);
+	return _kcapi_common_send_data(handle, iov, iovlen, MSG_MORE);
 }
 
 /**
@@ -1105,7 +1153,8 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	iov[0].iov_len = outlen;
 	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
 #endif
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT, 0);
+	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
+				      MSG_MORE);
 	if (0 > ret)
 		return ret;
 	if (assoc && handle->aead.assoclen) {
@@ -1115,17 +1164,17 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 		iov[1].iov_base = (void*)(uintptr_t)in;
 		iov[1].iov_len = inlen;
 		len += inlen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2);
+		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2, len, 0);
 	} else {
 		iov[0].iov_base = (void*)(uintptr_t)in;
 		iov[0].iov_len = inlen;
 		len = inlen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 1);
+		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 1, len, 0);
 	}
 	if (0 > ret)
 		return ret;
 
-	ret = _kcapi_common_splice_data(handle, len, out, outlen);
+	ret = _kcapi_common_read_data(handle, out, outlen);
 	if ((ret < (ssize_t)handle->aead.taglen))
 		return -E2BIG;
 
@@ -1250,7 +1299,8 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 	iov[0].iov_len = outlen;
 	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
 #endif
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT, 0);
+	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT,
+				      MSG_MORE);
 	if (0 > ret)
 		return ret;
 	if (assoc && handle->aead.assoclen) {
@@ -1263,7 +1313,7 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 		iov[2].iov_base = (void*)(uintptr_t)tag;
 		iov[2].iov_len = handle->aead.taglen;
 		len += handle->aead.taglen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 3);
+		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 3, len, 0);
 	} else {
 		iov[0].iov_base = (void*)(uintptr_t)in;
 		iov[0].iov_len = inlen;
@@ -1271,12 +1321,12 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 		iov[1].iov_base = (void*)(uintptr_t)tag;
 		iov[1].iov_len = handle->aead.taglen;
 		len += handle->aead.taglen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2);
+		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2, len, 0);
 	}
 	if (0 > ret)
 		return ret;
 
-	return _kcapi_common_splice_data(handle, len, out, outlen);
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 
 /**
@@ -1315,7 +1365,8 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 ssize_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 				   struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT, 1);
+	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_ENCRYPT,
+				       MSG_MORE);
 }
 
 /**
@@ -1354,7 +1405,8 @@ ssize_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 ssize_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 				   struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT, 1);
+	return _kcapi_common_send_meta(handle, iov, iovlen, ALG_OP_DECRYPT,
+				       MSG_MORE);
 }
 
 /**
@@ -1389,7 +1441,7 @@ ssize_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 ssize_t kcapi_aead_stream_update(struct kcapi_handle *handle,
 				 struct iovec *iov, size_t iovlen)
 {
-	return _kcapi_common_send_data(handle, iov, iovlen, 1);
+	return _kcapi_common_send_data(handle, iov, iovlen, MSG_MORE);
 }
 
 /**
@@ -1623,9 +1675,7 @@ ssize_t kcapi_md_update(struct kcapi_handle *handle,
 static ssize_t _kcapi_md_final(struct kcapi_handle *handle,
 			       unsigned char *buffer, size_t len)
 {
-	ssize_t r;
 	struct iovec iov;
-	struct msghdr msg;
 
 	if (len < (unsigned long)handle->info.hash_digestsize) {
 		fprintf(stderr,
@@ -1637,20 +1687,7 @@ static ssize_t _kcapi_md_final(struct kcapi_handle *handle,
 
 	iov.iov_base = (void*)(uintptr_t)buffer;
 	iov.iov_len = len;
-	msg.msg_name = NULL;
-	msg.msg_namelen = 0;
-	msg.msg_iov = &iov;
-	msg.msg_iovlen = 1;
-	msg.msg_control = NULL;
-	msg.msg_controllen = 0;
-
-	r = recvmsg(handle->opfd, &msg, 0);
-	if (r < 0)
-		return -EIO;
-	if (msg.msg_flags & MSG_TRUNC)
-		return -ENOMEM;
-
-	return r;
+	return _kcapi_common_recv_data(handle, &iov, 1);
 }
 
 /**
@@ -1697,6 +1734,7 @@ ssize_t kcapi_md_digest(struct kcapi_handle *handle,
 {
 	struct iovec iov;
 	ssize_t ret = 0;
+	size_t processed = 0;
 
 	if (!out || !outlen) {
 		fprintf(stderr,
@@ -1713,7 +1751,8 @@ ssize_t kcapi_md_digest(struct kcapi_handle *handle,
 	}
 
 	/* zero buffer length cannot be handled via splice */
-	if(!inlen) {
+	/* TODO check that heuristic for sendmsg is appropriate */
+	if(inlen == 0 /* < (1<<15) */) {
 		if (_kcapi_md_update(handle, in, inlen))
 			return -EIO;
 		return _kcapi_md_final(handle, out, outlen);
@@ -1721,12 +1760,20 @@ ssize_t kcapi_md_digest(struct kcapi_handle *handle,
 
 	/* normal zero copy */
 	iov.iov_base = (void*)(uintptr_t)in;
-	iov.iov_len = inlen;
 
-	ret = _kcapi_common_vmsplice_data(handle, &iov, 1);
-	if (0 > ret)
-		return ret;
-	return _kcapi_common_splice_data(handle, inlen, out, outlen);
+	while (inlen) {
+		size_t datalen = (inlen > MAXPIPELEN) ? MAXPIPELEN : inlen;
+
+		iov.iov_len = datalen;
+		ret = _kcapi_common_vmsplice_data(handle, &iov, 1, datalen,
+						  SPLICE_F_MORE);
+		if (0 > ret)
+			return ret;
+		processed += ret;
+		iov.iov_base = (void*)(uintptr_t)(in + processed);
+		inlen -= ret;
+	}
+	return _kcapi_common_read_data(handle, out, outlen);
 }
 /**
  * kcapi_md_digestsize() - return the size of the message digest
@@ -1801,11 +1848,7 @@ int kcapi_rng_destroy(struct kcapi_handle *handle)
 int kcapi_rng_seed(struct kcapi_handle *handle, unsigned char *seed,
 		   size_t seedlen)
 {
-	if (setsockopt(handle->tfmfd, SOL_ALG, ALG_SET_RNG_SEED,
-		       seed, seedlen) == -1)
-		return -EINVAL;
-
-	return 0;
+	return _kcapi_common_setkey(handle, seed, seedlen);
 }
 
 /**
@@ -1822,23 +1865,15 @@ ssize_t kcapi_rng_generate(struct kcapi_handle *handle,
 {
 	ssize_t out = 0;
 	struct iovec iov;
-	struct msghdr msg;
 
 	while (len) {
 		ssize_t r = 0;
 
 		iov.iov_base = (void*)(uintptr_t)buffer;
 		iov.iov_len = len;
-		msg.msg_name = NULL;
-		msg.msg_namelen = 0;
-		msg.msg_iov = &iov;
-		msg.msg_iovlen = 1;
-		msg.msg_control = NULL;
-		msg.msg_controllen = 0;
-
-		r = recvmsg(handle->opfd, &msg, 0);
+		r = _kcapi_common_recv_data(handle, &iov, 1);
 		if (0 >= r)
-			return -EIO;
+			return r;
 		len -= r;
 		out += r;
 
